# SPDX-FileCopyrightText: 2023 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Top-level package for Code128 library's python extension."""

__author__ = "Gabriele Pongelli"
__email__ = "gabriele.pongelli@gmail.com"
__version__ = "0.4.0"
